package cerebral;

import java.awt.Color;
import java.awt.Font;
import java.awt.font.FontRenderContext;
import java.awt.font.TextLayout;
import java.awt.geom.Point2D;
import java.awt.geom.Rectangle2D;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import prefuse.Display;
import prefuse.Visualization;
import prefuse.data.Edge;
import prefuse.data.Node;
import prefuse.data.Tuple;
import prefuse.data.expression.AndPredicate;
import prefuse.data.expression.BooleanLiteral;
import prefuse.data.expression.ColumnExpression;
import prefuse.data.expression.ComparisonPredicate;
import prefuse.data.expression.NumericLiteral;
import prefuse.data.tuple.TupleSet;
import prefuse.render.Renderer;
import prefuse.util.FontLib;
import prefuse.util.GraphicsLib;
import prefuse.util.PrefuseLib;
import prefuse.util.display.DisplayLib;
import prefuse.visual.EdgeItem;
import prefuse.visual.NodeItem;
import prefuse.visual.VisualItem;
import cerebral.SmallMultipleContainer.ViewAndTitleContainer;
import cerebral.layout.SearchGridLayout;
import cerebral.parallelCoordinates.ParallelCoordinates;

public class ViewCoordinator implements PropertyChangeListener {
    private List views;

    private Point2D down;

    private Cerebral mainDisplay;

    private SmallMultipleView select1;

    private String hoverNode;

    private ParallelCoordinates PC;

    private SmallMultipleContainer smc;

    public ViewCoordinator(Display mainDisplay, ParallelCoordinates PC) {
        views = new ArrayList();
        views.add(mainDisplay.getVisualization());
        down = new Point2D.Double();
        this.mainDisplay = (Cerebral) mainDisplay;
        ControlPanel.instance().addPropertyChangeListener(this);
        PC.addPropertyChangeListener(this);
        this.PC = PC;
        smc = null;
    }

    public void addView(SmallMultipleView view) {
        views.add(view.getVisualization());
    }

    public void updateNodePosition(NodeItem node, boolean updateCurX) {

        for (int i = 0; i < views.size(); i++) {
            Visualization vis = (Visualization) views.get(i);
            NodeItem replicate = (NodeItem) vis.getVisualItem(Cerebral.NODES,
                    node.getSourceTuple());
            if (replicate != null) {
                replicate.setEndX(node.getEndX());
                replicate.setEndY(node.getEndY());
                if (updateCurX) {
                    replicate.setX(node.getEndX());
                    replicate.setY(node.getEndY());
                }
            }
        }
    }

    public void updateNodePositions(Visualization sourceVis) {

        Iterator nodesIt = sourceVis.getVisualGroup(Cerebral.NODES).tuples();
        while (nodesIt.hasNext()) {
            NodeItem node = (NodeItem) nodesIt.next();
            for (int i = 0; i < views.size(); i++) {
                Visualization vis = (Visualization) views.get(i);
                NodeItem replicate = (NodeItem) vis.getVisualItem(
                        Cerebral.NODES, node.getSourceTuple());
                if (replicate != null) {
                    PrefuseLib.setX(replicate, node, node.getX());
                    PrefuseLib.setY(replicate, node, node.getY());
                }
            }
        }
        repaintAll();

    }

    public void updateEdgeAttributes(Visualization sourceVis, String attrib) {
        Iterator edgesIt = sourceVis.getVisualGroup(Cerebral.EDGES).tuples();
        while (edgesIt.hasNext()) {
            EdgeItem edge = (EdgeItem) edgesIt.next();
            for (int i = 0; i < views.size(); i++) {
                Visualization vis = (Visualization) views.get(i);
                EdgeItem replicate = (EdgeItem) vis.getVisualItem(
                        Cerebral.EDGES, edge.getSourceTuple());
                if (replicate != null) {
                    replicate.set(attrib, edge.get(attrib));
                }
            }
        }
        repaintAll();
    }

    public void updateVisible(VisualItem item) {
        String group = item instanceof NodeItem ? Cerebral.NODES
                : Cerebral.EDGES;
        for (int i = 0; i < views.size(); i++) {
            Visualization vis = (Visualization) views.get(i);
            VisualItem replicate = vis.getVisualItem(group, item
                    .getSourceTuple());
            if (replicate != null) {
                replicate.setVisible(item.isVisible());
            }
        }
        if (item instanceof NodeItem) {
            PC.updateVisible(item.getString(Cerebral.STRING_ID), item
                    .isVisible());
        }
    }

    public void repaintAll() {
        for (int i = 0; i < views.size(); i++) {
            Visualization vis = (Visualization) views.get(i);
            vis.repaint();
        }
    }

    public void fitViewToBounds(Rectangle2D bounds, int duration) {

        for (int i = 0; i < views.size(); i++) {
            Visualization vis = (Visualization) views.get(i);
            Display display = vis.getDisplay(0);
            if (!display.isTranformInProgress()) {
                DisplayLib.fitViewToBounds(display, bounds, duration);
            }
        }
    }

    private double getFitScale(Rectangle2D bounds, Display d) {
        double wb = Math.max(bounds.getCenterX() - bounds.getMinX(), bounds
                .getMaxX()
                - bounds.getCenterX());
        double hb = Math.max(bounds.getCenterY() - bounds.getMinY(), bounds
                .getMaxY()
                - bounds.getCenterY());
        return Math.min(d.getWidth() / (2 * wb), d.getHeight() / (2 * hb));

    }

    public void fitViewToBounds(long duration) {
        Rectangle2D bounds = Cerebral.getEndBounds(mainDisplay
                .getVisualization(), Cerebral.NODES);
        Display d = mainDisplay.getVisualization().getDisplay(0);
        double scale = getFitScale(bounds, d);
        double difference = 0.0;
        Rectangle2D labelRect = new Rectangle2D.Double();
        Rectangle2D boundsWLabels;
        int loopCount = 0;
        do {
            loopCount++;
            boundsWLabels = (Rectangle2D) bounds.clone();
            Iterator labelsIt = d.getVisualization().getGroup(
                    Cerebral.LINE_DIVIDER_LABELS).tuples();
            int fontSize = (int) (14 * (1 / scale));
            Font font = FontLib.getFont("Verdana", fontSize);
            FontRenderContext frc = Renderer.DEFAULT_GRAPHICS
                    .getFontRenderContext();

            while (labelsIt.hasNext()) {
                VisualItem labelItem = (VisualItem) labelsIt.next();
                String label = labelItem.getString(Cerebral.LABEL);
                if (label.length() > 0) {
                    TextLayout textLayout = new TextLayout(label, font, frc);
                    Rectangle2D textRect = textLayout.getBounds();
                    labelRect.setRect(labelItem.getX(), labelItem.getY(),
                            textRect.getWidth(), textRect.getHeight());

                    Rectangle2D.union(boundsWLabels, labelRect, boundsWLabels);
                }
            }
            double scaleWLabels = getFitScale(boundsWLabels, d);
            difference = scale - scaleWLabels;
            scale = scaleWLabels;
        } while (Math.abs(difference) > 0.000001 && loopCount < 10);
        GraphicsLib.expand(boundsWLabels, 50 + (int) 1 / scale);
        for (int i = 0; i < views.size(); i++) {
            Visualization vis = (Visualization) views.get(i);
            Display display = vis.getDisplay(0);
            if (!display.isTranformInProgress()) {
                DisplayLib.fitViewToBounds(display, boundsWLabels, 0);
            }
        }

        repaintAll();
    }

    public void pan(double x, double y) {
        for (int i = 0; i < views.size(); i++) {
            Visualization vis = (Visualization) views.get(i);
            Display display = vis.getDisplay(0);
            display.panAbs(x, y);
            display.repaint();
        }
    }

    public void zoom(double zoom) {
        for (int i = 0; i < views.size(); i++) {
            Visualization vis = (Visualization) views.get(i);
            Display display = vis.getDisplay(0);
            down.setLocation(display.getWidth() / 2, (display.getHeight()) / 2);
            display.zoom(down, zoom);
        }
        repaintAll();
    }

    public void setBackgroundColor(Color backgroundColor) {
        for (int i = 0; i < views.size(); i++) {
            Visualization vis = (Visualization) views.get(i);
            Display display = vis.getDisplay(0);
            display.setBackground(backgroundColor);
            display.repaint();
        }
    }

    public void select(String groupName, Node node, boolean selecting) {
        for (int i = 0; i < views.size(); i++) {
            Visualization vis = (Visualization) views.get(i);
            TupleSet selectedItems = vis.getFocusGroup(groupName);
            if (selectedItems != null) {
                NodeItem replicate = (NodeItem) vis.getVisualItem(
                        Cerebral.NODES, node);
                if (replicate != null) {
                    if (selecting) {
                        selectedItems.addTuple(replicate);
                    } else {
                        selectedItems.removeTuple(replicate);
                    }
                }
            }
        }
    }

    public void select(String groupName, Edge edge, boolean selecting) {
        for (int i = 0; i < views.size(); i++) {
            Visualization vis = (Visualization) views.get(i);
            TupleSet selectedItems = vis.getFocusGroup(groupName);
            if (selectedItems != null) {
                EdgeItem replicate = (EdgeItem) vis.getVisualItem(
                        Cerebral.EDGES, edge);
                if (replicate != null) {
                    if (selecting) {
                        selectedItems.addTuple(replicate);
                    } else {
                        selectedItems.removeTuple(replicate);
                    }
                }
            }
        }
    }

    public void clearFocusGroup(String groupName) {
        for (int i = 0; i < views.size(); i++) {
            Visualization vis = (Visualization) views.get(i);
            TupleSet selectedItems = vis.getFocusGroup(groupName);
            if (selectedItems != null) {
                selectedItems.clear();
            }
        }
    }

    public void run(String action) {
        for (int i = 0; i < views.size(); i++) {
            Visualization vis = (Visualization) views.get(i);
            vis.run(action);
        }
    }

    public void runAfter(String action, long delay) {
        for (int i = 0; i < views.size(); i++) {
            Visualization vis = (Visualization) views.get(i);
            vis.runAfter(action, delay);
        }
    }

    public void runAfter(String before, String after) {
        for (int i = 0; i < views.size(); i++) {
            Visualization vis = (Visualization) views.get(i);
            vis.runAfter(before, after);
        }
    }

    public void setShape(Tuple node, int shape) {
        for (int i = 0; i < views.size(); i++) {
            Visualization vis = (Visualization) views.get(i);
            NodeItem replicate = (NodeItem) vis.getVisualItem(Cerebral.NODES,
                    node);
            if (replicate != null) {
                replicate.setShape(shape);
            }

        }
    }

    public void enableSmallLineDivider(boolean showSeparators) {
        for (int i = 0; i < views.size(); i++) {
            Visualization vis = (Visualization) views.get(i);
            Iterator it = vis.getVisualGroup(Cerebral.LINE_DIVIDER).tuples();
            while (it.hasNext()) {
                VisualItem divider = (VisualItem) it.next();
                divider.setVisible(showSeparators);
            }

        }
        Iterator it = mainDisplay.getVisualization().getVisualGroup(
                Cerebral.LINE_DIVIDER_LABELS).tuples();
        while (it.hasNext()) {
            VisualItem label = (VisualItem) it.next();
            label.setVisible(showSeparators);
        }
        repaintAll();

    }

    public void positionDividers() {
        TupleSet dividers = mainDisplay.getVisualization().getVisualGroup(
                Cerebral.LINE_DIVIDER);
        double[] origBounds = new double[dividers.getTupleCount()];
        double[] layerBounds = new double[dividers.getTupleCount()];
        // load current positions as a starting point (useful if the
        // user pins every node in a layer)
        Iterator it = mainDisplay.getVisualization().getVisualGroup(
                Cerebral.LINE_DIVIDER).tuples();

        while (it.hasNext()) {
            VisualItem divider = (VisualItem) it.next();
            origBounds[divider.getRow()] = divider.getY();
        }

        int i = 0;
        it = dividers.tuples();
        while (it.hasNext()) {
            VisualItem divider = (VisualItem) it.next();
            int layer = divider.getRow();
            Iterator layerNodesIt = mainDisplay.getVisualization()
                    .getVisualGroup(Cerebral.NODES).tuples(
                            new AndPredicate(new AndPredicate(new AndPredicate(
                                    new ComparisonPredicate(
                                            ComparisonPredicate.EQ,
                                            new ColumnExpression(
                                                    Cerebral.BIO_CATEGORY),
                                            new NumericLiteral(layer)),
                                    new ComparisonPredicate(
                                            ComparisonPredicate.EQ,
                                            new ColumnExpression(
                                                    Cerebral.ANY_LAYER),
                                            new BooleanLiteral(false))),
                                    new ComparisonPredicate(
                                            ComparisonPredicate.EQ,
                                            new ColumnExpression(
                                                    Cerebral.PINNED),
                                            new BooleanLiteral(false))),
                                    new ComparisonPredicate(
                                            ComparisonPredicate.EQ,
                                            new ColumnExpression(
                                                    Cerebral.DOWNSTREAM_GENE),
                                            new BooleanLiteral(false))));

            boolean boundModified = false;
            layerBounds[i] = 0;
            while (layerNodesIt.hasNext()) {
                NodeItem nodeItem = (NodeItem) layerNodesIt.next();
                if (nodeItem.getEndY() > layerBounds[i]) {
                    layerBounds[i] = nodeItem.getEndY();
                    boundModified = true;
                }
            }
            if (boundModified) {
                layerBounds[i] += SearchGridLayout.SCALE / 2;
            } else {
                layerBounds[i] = origBounds[i];
            }
            i++;
        }

        setDividers(layerBounds);
        mainDisplay.setLayerLabelPositions();
    }

    private void setDividers(double[] layerBounds) {
        for (int i = 0; i < views.size(); i++) {
            Visualization vis = (Visualization) views.get(i);
            Iterator it = vis.getVisualGroup(Cerebral.LINE_DIVIDER).tuples();
            while (it.hasNext()) {
                VisualItem divider = (VisualItem) it.next();
                divider.setY(layerBounds[divider.getRow()]);
            }
        }
    }

    public void updateDividers() {
        TupleSet dividers = mainDisplay.getVisualization().getVisualGroup(
                Cerebral.LINE_DIVIDER);
        double[] origBounds = new double[dividers.getTupleCount()];
        Iterator it = mainDisplay.getVisualization().getVisualGroup(
                Cerebral.LINE_DIVIDER).tuples();

        while (it.hasNext()) {
            VisualItem divider = (VisualItem) it.next();
            origBounds[divider.getRow()] = divider.getY();
        }
        setDividers(origBounds);
        mainDisplay.setLayerLabelPositions();
        repaintAll();
    }

    public boolean userControlledColor() {
        return (select1 == null);
    }

    public void setQuality(boolean high) {
        for (int i = 0; i < views.size(); i++) {
            Visualization vis = (Visualization) views.get(i);
            Display d = vis.getDisplay(0);
            d.setHighQuality(high);
            d.repaint();
        }
    }

    public void setComparison(SmallMultipleView select1,
            SmallMultipleView select2) {
        if (select1 != null && select2 != null) {
            mainDisplay.colorManager.setCytoscapeFill(select1
                    .getComparisonFill(select2));
            mainDisplay.setMainLabel("Change from " + select1.getConditonName()
                    + " to " + select2.getConditonName());
            PC.setPrimarySelectedAxis(select1.getConditonName());
            PC.setSecondarySelectedAxis(select2.getConditonName());
        } else if (select1 != null && select2 == null) {
            mainDisplay.colorManager.setCytoscapeFill(select1.getFillColor());
            mainDisplay.setMainLabel(select1.getConditonName());
            PC.setPrimarySelectedAxis(select1.getConditonName());
            PC.setSecondarySelectedAxis(null);
        } else {
            mainDisplay.colorManager.setCytoscapeFill(null);
            mainDisplay.setMainLabel("");
            PC.setPrimarySelectedAxis(null);
            PC.setSecondarySelectedAxis(null);
        }
        mainDisplay.getVisualization().run("color");
        mainDisplay.getVisualization().run("animate");

    }

    public void setHover(String nodeLabel, boolean b) {
        NodeItem hoverNode = mainDisplay.getNodeItemFromCytoID(nodeLabel);
        hoverNode.setHover(b);
        for (int i = 0; i < views.size(); i++) {
            Visualization vis = (Visualization) views.get(i);
            NodeItem replicate = (NodeItem) vis.getVisualItem(Cerebral.NODES,
                    hoverNode.getSourceTuple());
            if (replicate != null) {
                replicate.setHover(b);
            }
        }
        PC.setHover(nodeLabel, b);
        run("color");
        run("animate");
    }

    public void setIgnore(String nodeLabel, boolean b) {
        NodeItem mainNode = mainDisplay.getNodeItemFromCytoID(nodeLabel);
        mainNode.setBoolean(Cerebral.IGNORE, b);
        for (int i = 0; i < views.size(); i++) {
            Visualization vis = (Visualization) views.get(i);
            NodeItem replicate = (NodeItem) vis.getVisualItem(Cerebral.NODES,
                    mainNode.getSourceTuple());
            if (replicate != null) {
                replicate.setBoolean(Cerebral.IGNORE, b);
            }
        }
    }

    public void removeSmallMultiples() {
        views.clear();
        views.add(mainDisplay.getVisualization());
    }

    public void propertyChange(PropertyChangeEvent e) {
        if (e.getPropertyName() == "quality") {
            Boolean state = (Boolean) e.getNewValue();
            setQuality(state.booleanValue());
        } else if (e.getPropertyName() == "hover") {
            if (e.getNewValue() == null) {
                setHover(hoverNode, false);
            } else {
                hoverNode = (String) e.getNewValue();
                setHover(hoverNode, true);
            }
        } else if (e.getPropertyName() == "ignore") {
            setIgnore((String) e.getOldValue(), e.getNewValue() == "t");
        } else if (e.getPropertyName() == "axisReorder") {
            String[] locations = ((String) e.getNewValue()).split(":");
            String axis = (String) e.getOldValue();
            if (smc != null) {
                smc.moveView(axis, Integer.parseInt(locations[0]), Integer
                        .parseInt(locations[1]));
            }
        } else if (e.getPropertyName() == "axisSelect"){            
            if (smc != null){
                ViewAndTitleContainer vtc=smc.findView((String)e.getOldValue());
                smc.viewSelected(vtc.viewAndTitle,  ((String)e.getNewValue()).compareTo("y")==0);

            }
        }

    }

    public void setSmallMultipleContainer(SmallMultipleContainer container) {
        smc = container;

    }

    public void viewMoved(int sourceLoc, int destLoc) {
        PC.axisMoved(sourceLoc, destLoc);
    }
}
